from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Iterable, List, Mapping, Optional, Tuple, TypeVar, cast

import bs4
import requests
from charmonium.cache import MemoizedGroup, memoize
from tqdm import tqdm

DEFAULT_PARSER = "html5lib"
group = MemoizedGroup(fine_grain_persistence=True, size="10MiB")


@dataclass
class CodeRecord:
    """Information about a code representing one entry in <https://ascl.net/code/all>."""

    ascl_id: Optional[Tuple[int, int]]
    title: str
    credit: List[str]
    abstract: str
    details_url: str

    def get_details(self) -> DetailedCodeRecord:
        return cast(DetailedCodeRecord, scrape_details(self.details_url))


def parse_ascl_id(ascl_id_str: str) -> Optional[Tuple[int, int]]:
    if m := re.match(r"\[ascl:(\d+).(\d+)\]", ascl_id_str):
        return (int(m.group(1)), int(m.group(2)))
    else:
        return None


_T = TypeVar("_T")


def unwrap(obj: Optional[_T]) -> _T:
    if obj is None:
        raise ValueError("Unable to parse page")
    return obj


@memoize(group=group)
def scrape_index(
    max_count: Optional[int] = None,
    verbose: bool = True,
) -> List[CodeRecord]:
    """Get `max_count` entries from <https://ascl.net/code/all>.

    Pass `None` to get all entries.

    This function caches the result, so if it is called with the same `max_count`, the result can be loaded from disk.
    """
    return list(scrape_index_lazy(max_count, verbose))


def scrape_index_lazy(
    max_count: Optional[int] = None,
    verbose: bool = True,
) -> Iterable[CodeRecord]:
    max_count = max_count if max_count is not None else 300000
    # 300000 ~ 100 * current value on 2022-03-10
    response = requests.get(f"https://ascl.net/code/all/limit/{max_count}")
    soup = bs4.BeautifulSoup(response.text, DEFAULT_PARSER)
    items = list(soup.select("div.codelist div.item"))
    u = unwrap
    item: bs4.Tag
    for item in tqdm(iter(items), total=len(items), disable=not verbose):
        assert item
        yield CodeRecord(
            ascl_id=parse_ascl_id(u(item.select_one("span.ascl_id")).text),
            title=u(item.select_one("span.title")).text.strip(),
            credit=[
                child.text for child in u(item.select_one("div.credit a")).children
            ],
            abstract=str(u(item.select_one("p"))),
            details_url=(
                "https://ascl.net/" + u(item.select_one("span.title a")).attrs["href"]
            ).replace(".net//", ".net/"),
        )


github_regex = re.compile(r"https?://github.com/[a-zA-Z0-9\.\-/]")


@dataclass
class DetailedCodeRecord:
    """Detailed information about a code, for example <https://ascl.net/0000.000>."""

    ascl_id: Optional[Tuple[int, int]]
    title: str
    credit: List[str]
    abstract: str
    url: str
    code_sites: List[str]
    used_in: List[str]
    described_in: List[str]
    bibcode: Optional[str]
    preferred_citation_method: Optional[str]
    discuss_url: str
    views: int

    @property
    def github(self) -> Optional[str]:
        return cast(Optional[str], get_github_for(self))


@memoize(group=group)
def get_github_for(record: DetailedCodeRecord) -> Optional[str]:
    # First, see if any code_site is a github site.
    for site in record.code_sites:
        if github_regex.match(site):
            return site

    # Second, see if any code_site links to a github site.
    for site in record.code_sites:
        try:
            # A lot of old sites take forever to time out.
            text = requests.get(site, timeout=4).text
        except requests.exceptions.RequestException:
            # A lot of old sites are dead.
            continue
        for tag in bs4.BeautifulSoup(text, DEFAULT_PARSER).find_all("a"):
            if "href" in tag.attrs and github_regex.match(tag.attrs["href"]):
                return cast(str, tag.attrs["href"])

    # Third, give up.
    return None


def dl_to_dict(dl: bs4.Tag) -> Mapping[str, bs4.Tag]:
    return {
        key.text: cast(bs4.Tag, val)
        for key, val in zip(dl.find_all("dt"), dl.find_all("dd"))
    }


@memoize(group=group)
def scrape_details(
    url: str,
) -> DetailedCodeRecord:
    """Get detailed information about a code."""
    response = requests.get(url)
    soup = bs4.BeautifulSoup(response.text, DEFAULT_PARSER)
    item = soup.select_one("div.codelist div.item")
    assert item
    sites_soup = item.select_one("dl.sites")
    sites = dl_to_dict(sites_soup) if sites_soup else {}
    bibcode = item.select_one("dl.sites.bibcode > dd")
    cite_method = item.select_one("div.cite_method")
    u = unwrap
    return DetailedCodeRecord(
        ascl_id=parse_ascl_id(u(item.select_one("span.ascl_id")).text),
        title=u(item.select_one("span.title")).text.strip(),
        credit=[child.text for child in u(item.select_one("div.credit a")).children],
        abstract=str(item.select_one("p")),
        url=url,
        code_sites=(
            [link.attrs["href"] for link in sites["Code site:"].select("a")]
            if "Code site:" in sites
            else []
        ),
        used_in=(
            [link.attrs["href"] for link in sites["Used in:"].select("a")]
            if "Used in:" in sites
            else []
        ),
        described_in=(
            [link.attrs["href"] for link in sites["Described in:"].select("a")]
            if "Described in:" in sites
            else []
        ),
        bibcode=bibcode.text if bibcode else None,
        preferred_citation_method=str(cite_method.select_one("p"))
        if cite_method
        else None,
        discuss_url=u(item.select_one("div.discuss > a")).attrs["href"],
        views=int(u(item.select_one("div.views")).text[7:]),
    )



from .lib import scrape_index

# __name__ == "__main__" is needed so pytest ignores this.
if __name__ == "__main__":
    results = scrape_index(100)
    for result in results:
        result.get_details()



__version__ = "0.3.3"
from .lib import CodeRecord as CodeRecord
from .lib import DetailedCodeRecord as DetailedCodeRecord
from .lib import scrape_details as scrape_details
from .lib import scrape_index as scrape_index



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ascl_net_scraper']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2',
 'charmonium.cache>=1.2.6,<2.0.0',
 'html5lib>=1.1,<2.0',
 'requests>=2.27.1,<3.0.0']

setup_kwargs = {
    'name': 'ascl-net-scraper',
    'version': '0.3.3',
    'description': 'Scrapes the data from https://ascl.net',
    'long_description': '==========================\nascl_net_scraper\n==========================\n\n.. image:: https://img.shields.io/pypi/v/ascl_net_scraper\n   :alt: PyPI Package\n   :target: https://pypi.org/project/ascl_net_scraper\n.. image:: https://img.shields.io/pypi/dm/ascl_net_scraper\n   :alt: PyPI Downloads\n   :target: https://pypi.org/project/ascl_net_scraper\n.. image:: https://img.shields.io/pypi/l/ascl_net_scraper\n   :alt: PyPI License\n.. image:: https://img.shields.io/pypi/pyversions/ascl_net_scraper\n   :alt: Python Versions\n.. image:: https://img.shields.io/github/stars/charmoniumQ/ascl_net_scraper?style=social\n   :alt: GitHub stars\n   :target: https://github.com/charmoniumQ/ascl_net_scraper\n.. image:: https://github.com/charmoniumQ/ascl_net_scraper/actions/workflows/main.yaml/badge.svg\n   :alt: CI status\n   :target: https://github.com/charmoniumQ/ascl_net_scraper/actions/workflows/main.yaml\n.. image:: https://img.shields.io/github/last-commit/charmoniumQ/charmonium.determ_hash\n   :alt: GitHub last commit\n   :target: https://github.com/charmoniumQ/ascl_net_scraper/commits\n.. image:: https://img.shields.io/librariesio/sourcerank/pypi/ascl_net_scraper\n   :alt: libraries.io sourcerank\n   :target: https://libraries.io/pypi/ascl_net_scraper\n.. image:: https://img.shields.io/badge/docs-yes-success\n   :alt: Documentation link\n.. image:: http://www.mypy-lang.org/static/mypy_badge.svg\n   :target: https://mypy.readthedocs.io/en/stable/\n   :alt: Checked with Mypy\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Code style: black\n\nScrapes the data from https://ascl.net\n\n----------\nQuickstart\n----------\n\nIf you don\'t have ``pip`` installed, see the `pip install\nguide`_.\n\n.. _`pip install guide`: https://pip.pypa.io/en/latest/installing/\n\n.. code-block:: console\n\n    $ pip install ascl_net_scraper\n\n>>> from rich.pretty import pprint # for pretty printing\n>>> import ascl_net_scraper\n>>> codes = ascl_net_scraper.scrape_index(5)\n>>> pprint(codes[0]) # doctest: +ELLIPSIS\nCodeRecord(\n│   ascl_id=None,\n│   title=\'2-DUST: Dust radiative transfer code\',\n│   credit=[\'Ueta, Toshiya\'],\n│   abstract=\'<p>...</p>\',\n│   details_url=\'https://ascl.net/1604.006\'\n)\n>>> pprint(codes[0].get_details(), max_string=70) # doctest: +ELLIPSIS\nDetailedCodeRecord(\n│   ascl_id=None,\n│   title=\'2-DUST: Dust radiative transfer code\',\n│   credit=[\'Ueta, Toshiya\'],\n│   abstract=\'<p>2-DUST is a general-purpose dust radiative transfer code for an axi\'+319,\n│   url=\'https://ascl.net/1604.006\',\n│   code_sites=[\'https://github.com/sundarjhu/2-DUST/\'],\n│   used_in=[\'https://ui.adsabs.harvard.edu/abs/2004ApJ...614..371M\'],\n│   described_in=[\'https://ui.adsabs.harvard.edu/abs/2003ApJ...586.1338U\'],\n│   bibcode=\'2016ascl.soft04006U\',\n│   preferred_citation_method=\'<p><a href="https://ui.adsabs.harvard.edu/abs/2003ApJ...586.1338U">htt\'+58,\n│   discuss_url=\'/phpBB3/viewtopic.php?t=33976\',\n│   views=...\n)\n',
    'author': 'Samuel Grayson',
    'author_email': 'sam+dev@samgrayson.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/charmoniumQ/ascl_net_scraper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
